from __future__ import print_function
import pickle, util, sys, tqdm, time, json, uuid
import os.path
from multiprocessing import Pool
from colorama import Fore, Style
from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.errors import HttpError

# If modifying these scopes, delete the file token.pickle.
SCOPES = ['https://www.googleapis.com/auth/drive']
PROCESS_COUNT = 5

def add_account():
    tokenfile = f'token_{str(uuid.uuid4())}.pickle'
    os.makedirs('tokens', exist_ok=True)
    flow = InstalledAppFlow.from_client_secrets_file(
                'credentials.json', SCOPES)
    creds = flow.run_local_server(port=0)
    # Save the credentials for the next run
    with open(f'tokens/{tokenfile}', 'wb') as token:
        pickle.dump(creds, token)

def migrate():
    # Migrate token.pickle file to tokens folder
    os.makedirs('tokens', exist_ok=True)
    if os.path.exists('token.pickle'):
        print("Older token configuration detected. Migrating to the new setup")
        new_name = f'token_{str(uuid.uuid4())}.pickle'
        os.rename('token.pickle', f'tokens/{new_name}')

def get_accounts():
    return [x for x in os.listdir('tokens') if x.endswith('.pickle')]

def get_service(tokenfile):
    """Shows basic usage of the Drive v3 API.
    Prints the names and ids of the first 10 files the user has access to.
    """
    creds = None
    # The file token.pickle stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    if os.path.exists(f'tokens/{tokenfile}'):
        with open(f'tokens/{tokenfile}', 'rb') as token:
            creds = pickle.load(token)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                'credentials.json', SCOPES)
            creds = flow.run_local_server(port=0)
        # Save the credentials for the next run
        with open(f'tokens/{tokenfile}', 'wb') as token:
            pickle.dump(creds, token)

    service = build('drive', 'v3', credentials=creds)
    return service

def download_helper(args):
    rlc = util.download(*args)
    return (args[1]['name'], rlc)

def mapped_dl(args):
    rlc = util.download(*args, noiter=True)
    return (args[1]['name'], rlc)

def main(console_call=True):
    # Set path
    if console_call:
        os.chdir(os.path.dirname(os.path.abspath(__file__)))
    else:
        os.chdir(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

    if not os.path.exists('credentials.json'):
        print(f"Missing Credentials!\nEnable Drive API by clicking the blue button at https://developers.google.com/drive/api/v3/quickstart/python \nDownload the credentials.json file and save it here: {os.getcwd()}")

    migrate()
    accounts = get_accounts()
    service = None
    search = False
    skip = False
    noiter = False

    # File Listing
    if len(sys.argv) < 2:
        print("Usage: tdlist <folderid> <destination>")
        sys.exit(-1)
    else:
        if sys.argv[1] == '--path':
            util.save_default_path(sys.argv[2])
            sys.exit(0)
        if sys.argv[1] == '--add':
            add_account()
            print("Account successfully added!")
            sys.exit(0)
        if '--search' in sys.argv:
            search = True
            sys.argv.remove('--search')
        if '--skip' in sys.argv:
            skip = True
            sys.argv.remove('--skip')
        if '--debug' in sys.argv:
            util.DEBUG = True
            sys.argv.remove('--debug')
        if '--noiter' in sys.argv:
            noiter = True
            sys.argv.remove('--noiter')
        if '--proc' in sys.argv:
            index = sys.argv.index('--proc')
            PROCESS_COUNT = int(sys.argv[index + 1])
            sys.argv.pop(index + 1)
            sys.argv.pop(index)
        folderid = util.get_folder_id(sys.argv[1])
        if len(sys.argv) > 2:
            destination = sys.argv[2]
        elif os.path.isfile('config.json'):
            with open('config.json', 'r') as f:
                config = json.load(f)
            destination = config['default_path']
        else:
            destination = os.getcwd()
        
    file_dest = []

    def build_files():
        try:
            kwargs = {'top': folderid, 'by_name': False}
            for path, root, dirs, files in util.walk(service, **kwargs):
                path = ["".join([c for c in dirname if c.isalpha() or c.isdigit() or c in [' ', '-', '_', '.', '(', ')', '[', ']']]).rstrip() for dirname in path]
                for f in files:
                    dest = os.path.join(destination, os.path.join(*path))
                    file_dest.append((service, f, dest, skip))
            if file_dest != []:
                # First valid account found, break to prevent further searches
                return True
        except ValueError: # mimetype is not a folder
            dlfile = service.files().get(fileId=folderid, supportsAllDrives=True).execute()
            print(f"\nNot a valid folder ID. \nDownloading the file : {dlfile['name']}")
            # Only use a single process for downloading 1 file
            util.download(service, dlfile, destination, skip)
            sys.exit(0)
        except HttpError:
            print(f"{Fore.RED}File not found in account: {acc}{Style.RESET_ALL}")
            return False

    searches = []
    for acc in accounts:
        service = get_service(acc)
        if not search:
            valid = build_files()
            if valid: break
        else:
            searches += util.querysearch(service, folderid, 'rand', True)

    if searches != []:
        print("\nEnter the number of the folder you want to download:")
        for i in range(len(searches)):
            print(f" {Fore.YELLOW}{str(i+1).ljust(3)}{Style.RESET_ALL}    {searches[i]['name']}")
        print(f"\n{Fore.GREEN}Index:{Style.RESET_ALL} ", end='')
        index = int(input()) - 1
        if index + 1 > len(searches):
            print(f"{Fore.RED}Invalid Index. Exiting.{Style.RESET_ALL}")
            sys.exit(1)
        folderid = searches[index]['id']
        build_files()

    if service == None:
        # No accounts found with access to the drive link, exit gracefully
        print("No valid accounts with access to the file/folder. Exiting...")
        sys.exit(1)
    try:
        p = Pool(PROCESS_COUNT)
        if noiter:
            p.map(mapped_dl, file_dest)
        else:
            pbar = tqdm.tqdm(p.imap_unordered(download_helper, file_dest), total=len(file_dest))
            start = time.time()
            for i in pbar:
                rlc = i[1]
                status, main_str, end_str = util.get_download_status(rlc, start)
                pbar.write(status + main_str + f' {i[0]}' + end_str)
            p.close()
            p.join()
        if util.DEBUG:
            util.debug_write(f'debug_{int(time.time())}.log')
    except ImportError:
        # Multiprocessing is not supported (example: Android Devices)
        for fd in file_dest:
            download_helper(fd)

if __name__ == '__main__':
    main(False)


from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from googleapiclient.http import MediaIoBaseDownload
from colorama import Fore, Style
import io, os, shutil, uuid, sys, json, time

FOLDER = 'application/vnd.google-apps.folder'
DEBUG = False
CHUNK_SIZE = 20 * 1024 * 1024 # 20MB chunks

DEBUG_STATEMENTS = [] # cache all debug statements

def debug_write(logfile):
    with open(logfile, 'w') as f:
        f.write('\n'.join(DEBUG_STATEMENTS))
    print(f"{Fore.YELLOW}DEBUG LOG SAVED HERE:{Style.RESET_ALL} {logfile}")

def list_td(service):
    # Call the Drive v3 API
    results = service.drives().list(pageSize=100).execute()

    if not results['drives']:
        return None
    else:
        return results['drives']

def iterfiles(service, name=None, is_folder=None, parent=None, order_by='folder,name,createdTime'):
    q = []
    if name is not None:
        q.append("name = '%s'" % name.replace("'", "\\'"))
    if is_folder is not None:
        q.append("mimeType %s '%s'" % ('=' if is_folder else '!=', FOLDER))
    if parent is not None:
        q.append("'%s' in parents" % parent.replace("'", "\\'"))
    params = {'pageToken': None, 'orderBy': order_by, 'includeItemsFromAllDrives': True, 'supportsAllDrives': True}
    if q:
        params['q'] = ' and '.join(q)
    while True:
        response = service.files().list(**params).execute()
        for f in response['files']:
            yield f
        try:
            params['pageToken'] = response['nextPageToken']
        except KeyError:
            return

def walk(service, top='root', by_name=False):
    if by_name:
        top, = iterfiles(service, name=top, is_folder=True)
    else:
        top = service.files().get(fileId=top, supportsAllDrives=True).execute()
        if top['mimeType'] != FOLDER:
            raise ValueError('not a folder: %r' % top)
    stack = [((top['name'],), top)]
    print(f"Indexing: {Fore.YELLOW}{top['name']}{Style.RESET_ALL}\nFolder ID: {Fore.YELLOW}{top['id']}{Style.RESET_ALL}\n")
    while stack:
        path, top = stack.pop()
        dirs, files = is_file = [], []
        for f in iterfiles(service, parent=top['id']):
            is_file[f['mimeType'] != FOLDER].append(f)
        yield path, top, dirs, files
        if dirs:
            stack.extend((path + (d['name'],), d) for d in reversed(dirs))

def querysearch(service, name=None, drive_id=None, is_folder=None, parent=None, order_by='folder,name,createdTime'):
    q = []
    items = []
    if name is not None:
        q.append("name contains '%s'" % name.replace("'", "\\'"))
    if is_folder is not None:
        q.append("mimeType %s '%s'" % ('=' if is_folder else '!=', FOLDER))
    if parent is not None:
        q.append("'%s' in parents" % parent.replace("'", "\\'"))
    if drive_id == None:
        params = {'pageToken': None, 'orderBy': order_by, 'includeItemsFromAllDrives': True, 'supportsAllDrives': True}
    else:
        params = {'pageToken': None, 'orderBy': order_by, 'includeItemsFromAllDrives': True, 'supportsAllDrives': True, 'corpora': 'allDrives'}
    if q:
        params['q'] = ' and '.join(q)
    while len(items) < 10:
        response = service.files().list(**params).execute()
        for f in response['files']:
            items.append(f)
        try:
            params['pageToken'] = response['nextPageToken']
        except KeyError:
            break
    return items

def download(service, file, destination, skip=False, noiter=False):
    # file is a dictionary with file id as well as name
    if skip and os.path.exists(os.path.join(destination, file['name'])):
        return -1
    mimeType = file['mimeType']
    if "application/vnd.google-apps" in mimeType:
        if "form" in mimeType: return -1
        elif "document" in mimeType:
            dlfile = service.files().export_media(fileId=file['id'], mimeType='application/vnd.openxmlformats-officedocument.wordprocessingml.document')
        elif "spreadsheet" in mimeType:
            dlfile = service.files().export_media(fileId=file['id'], mimeType='application/vnd.openxmlformats-officedocument.spreadsheetml.sheet')
        elif "presentation" in mimeType:
            dlfile = service.files().export_media(fileId=file['id'], mimeType='application/vnd.openxmlformats-officedocument.presentationml.presentation')
        else:
            dlfile = service.files().export_media(fileId=file['id'], mimeType='application/pdf')
    else:
        dlfile = service.files().get_media(fileId=file['id'], supportsAllDrives=True)
    rand_id = str(uuid.uuid4())
    os.makedirs('buffer', exist_ok=True)
    fh = io.FileIO(os.path.join('buffer', rand_id), 'wb')
    downloader = MediaIoBaseDownload(fh, dlfile, chunksize=CHUNK_SIZE)
    if noiter: print(f"{Fore.GREEN}Downloading{Style.RESET_ALL} {file['name']} ...")
    done = False
    rate_limit_count = 0
    while done is False and rate_limit_count < 20:
        try:
            status, done = downloader.next_chunk()
        except Exception as ex:
            DEBUG_STATEMENTS.append(f'File Name: {file["name"]}, File ID: {file["id"]}, Exception: {ex}')
            rate_limit_count += 1
    fh.close()
    if noiter and rate_limit_count == 20: print(f"{Fore.RED}Error      {Style.RESET_ALL} {file['name']} ...")
    os.makedirs(destination, exist_ok=True)
    while True:
        try:
            shutil.move(os.path.join('buffer', rand_id), os.path.join(destination, file['name']))
            break
        except PermissionError:
            # wait out the file write before attempting to move
            pass
    return rate_limit_count

def get_folder_id(link):
    # function to isolate folder id
    if 'drive.google.com' in link:
        link = link.split('/view')[0].split('/edit')[0] # extensions to file names
        link = link.rsplit('/', 1)[-1] # final backslash
        link = link.split('?usp')[0] # ignore usp=sharing and usp=edit
        # open?id=
        link = link.rsplit('open?id=')[-1] # only take what is after open?id=
        return link
    else:
        return link

def save_default_path(path):
    if os.path.isfile('config.json'):
        with open('config.json', 'r') as f:
            config = json.load(f)
        config['default_path'] = path
    else:
        config = {}
        config['default_path'] = path
    with open('config.json', 'w') as f:
        f.write(json.dumps(config, indent= 4))

def get_download_status(rlc, start):
    if rlc == -1: # skipped file
        status = f'{Fore.CYAN}Skipped:   {Style.RESET_ALL} '
    elif rlc == 0:
        status = f'{Fore.GREEN}Downloaded:{Style.RESET_ALL} '
    elif rlc < 20:
        status = f'{Fore.YELLOW}Warning:   {Style.RESET_ALL} '
    else:
        status = f'{Fore.RED}Error:     {Style.RESET_ALL} '
    time_req = str(int(time.time() - start)) + 's'
    main_str = f'{Fore.BLUE}[Time: {time_req.rjust(5)}]{Style.RESET_ALL}'
    end_str = ''
    if rlc > 0 and rlc < 20:
        end_str += f' [Rate Limit Count: {rlc}] File saved'
    elif rlc >= 20:
        end_str += f' [Rate Limit Count: {rlc}] Partial file saved'
    return (status, main_str, end_str)


from setuptools import setup, find_packages

package_name = "drivedl"

with open("README.md", "r") as f:
    readme = f.read()

setup(
    name="drivedl",
    version="1.1",
    description="Download files from Google Drive (inclusive of teamdrives, shared with me and my drive) concurrently.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/architdate/drivedl",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=[
        "google-api-python-client",
        "google-auth-httplib2",
        "google-auth-oauthlib",
        "tqdm",
        "colorama"
    ],
    license='MIT',
    zip_safe=False,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    author="Archit Date",
    author_email="architdate@gmail.com",
    entry_points={
        'console_scripts': [
            'drivedl=drivedl.drivedl:main',
        ],
    },
)